@@ -2,21 +2,44 @@ use swc_atoms::{js_word, JsWord};
 use swc_common::{
     collections::{AHashMap, AHashSet},
     sync::Lrc,
+    EqIgnoreSpan,
 };
 use swc_ecma_ast::*;
 use swc_ecma_transforms_base::perf::Parallel;
 use swc_ecma_transforms_macros::parallel;
-use swc_ecma_utils::{collect_decls, Id};
+use swc_ecma_utils::{collect_decls, ident::IdentLike, Id};
 use swc_ecma_visit::{as_folder, noop_visit_mut_type, Fold, VisitMut, VisitMutWith};
 
+/// The key will be compared using [EqIgnoreSpan::eq_ignore_span], and matched
+/// expressions will be replaced with the value.
+pub type GlobalExprMap = Lrc<Vec<(Expr, Expr)>>;
+
+/// Create a global inlining pass, which replaces expressions with the specified
+/// value.
 pub fn inline_globals(
     envs: Lrc<AHashMap<JsWord, Expr>>,
     globals: Lrc<AHashMap<JsWord, Expr>>,
     typeofs: Lrc<AHashMap<JsWord, JsWord>>,
+) -> impl Fold + VisitMut {
+    inline_globals2(envs, globals, Default::default(), typeofs)
+}
+
+/// Create a global inlining pass, which replaces expressions with the specified
+/// value.
+///
+/// See [GlobalExprMap] for description.
+///
+/// Note: Values specified in `global_exprs` have higher precedence than
+pub fn inline_globals2(
+    envs: Lrc<AHashMap<JsWord, Expr>>,
+    globals: Lrc<AHashMap<JsWord, Expr>>,
+    global_exprs: GlobalExprMap,
+    typeofs: Lrc<AHashMap<JsWord, JsWord>>,
 ) -> impl Fold + VisitMut {
     as_folder(InlineGlobals {
         envs,
         globals,
+        global_exprs,
         typeofs,
         bindings: Default::default(),
     })
@@ -26,6 +49,8 @@ pub fn inline_globals(
 struct InlineGlobals {
     envs: Lrc<AHashMap<JsWord, Expr>>,
     globals: Lrc<AHashMap<JsWord, Expr>>,
+    global_exprs: GlobalExprMap,
+
     typeofs: Lrc<AHashMap<JsWord, JsWord>>,
 
     bindings: Lrc<AHashSet<Id>>,
@@ -44,14 +69,28 @@ impl VisitMut for InlineGlobals {
     noop_visit_mut_type!();
 
     fn visit_mut_expr(&mut self, expr: &mut Expr) {
-        expr.visit_mut_children_with(self);
-
         match expr {
             Expr::Ident(Ident { ref sym, span, .. }) => {
                 if self.bindings.contains(&(sym.clone(), span.ctxt)) {
                     return;
                 }
+            }
 
+            _ => {}
+        }
+
+        for (key, value) in self.global_exprs.iter() {
+            if key.eq_ignore_span(&*expr) {
+                *expr = value.clone();
+                expr.visit_mut_with(self);
+                return;
+            }
+        }
+
+        expr.visit_mut_children_with(self);
+
+        match expr {
+            Expr::Ident(Ident { ref sym, .. }) => {
                 // It's ok because we don't recurse into member expressions.
                 if let Some(value) = self.globals.get(sym) {
                     let mut value = value.clone();
@@ -152,6 +191,30 @@ impl VisitMut for InlineGlobals {
         module.visit_mut_children_with(self);
     }
 
+    fn visit_mut_prop(&mut self, p: &mut Prop) {
+        p.visit_mut_children_with(self);
+
+        match p {
+            Prop::Shorthand(i) => {
+                if self.bindings.contains(&i.to_id()) {
+                    return;
+                }
+
+                // It's ok because we don't recurse into member expressions.
+                if let Some(mut value) = self.globals.get(&i.sym).cloned().map(Box::new) {
+                    value.visit_mut_with(self);
+                    *p = Prop::KeyValue(KeyValueProp {
+                        key: PropName::Ident(i.clone()),
+                        value,
+                    });
+                }
+
+                return;
+            }
+            _ => {}
+        }
+    }
+
     fn visit_mut_script(&mut self, script: &mut Script) {
         self.bindings = Lrc::new(collect_decls(&*script));
 
@@ -212,12 +275,7 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |tester| as_folder(InlineGlobals {
-            envs: envs(tester, &[]),
-            globals: globals(tester, &[]),
-            typeofs: Default::default(),
-            bindings: Default::default()
-        }),
+        |tester| inline_globals(envs(tester, &[]), globals(tester, &[]), Default::default(),),
         issue_215,
         r#"if (process.env.x === 'development') {}"#,
         r#"if (process.env.x === 'development') {}"#
@@ -225,12 +283,11 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |tester| as_folder(InlineGlobals {
-            envs: envs(tester, &[("NODE_ENV", "development")]),
-            globals: globals(tester, &[]),
-            typeofs: Default::default(),
-            bindings: Default::default()
-        }),
+        |tester| inline_globals(
+            envs(tester, &[("NODE_ENV", "development")]),
+            globals(tester, &[]),
+            Default::default(),
+        ),
         node_env,
         r#"if (process.env.NODE_ENV === 'development') {}"#,
         r#"if ('development' === 'development') {}"#
@@ -238,25 +295,23 @@ mod tests {
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |tester| as_folder(InlineGlobals {
-            envs: envs(tester, &[]),
-            globals: globals(tester, &[("__DEBUG__", "true")]),
-            typeofs: Default::default(),
-            bindings: Default::default()
-        }),
-        inline_globals,
+        |tester| inline_globals(
+            envs(tester, &[]),
+            globals(tester, &[("__DEBUG__", "true")]),
+            Default::default(),
+        ),
+        globals_simple,
         r#"if (__DEBUG__) {}"#,
         r#"if (true) {}"#
     );
 
     test!(
         ::swc_ecma_parser::Syntax::default(),
-        |tester| as_folder(InlineGlobals {
-            envs: envs(tester, &[]),
-            globals: globals(tester, &[("debug", "true")]),
-            typeofs: Default::default(),
-            bindings: Default::default()
-        }),
+        |tester| inline_globals(
+            envs(tester, &[]),
+            globals(tester, &[("debug", "true")]),
+            Default::default(),
+        ),
         non_global,
         r#"if (foo.debug) {}"#,
         r#"if (foo.debug) {}"#
@@ -264,12 +319,7 @@ mod tests {
 
     test!(
         Default::default(),
-        |tester| as_folder(InlineGlobals {
-            envs: envs(tester, &[]),
-            globals: globals(tester, &[]),
-            typeofs: Default::default(),
-            bindings: Default::default()
-        }),
+        |tester| inline_globals(envs(tester, &[]), globals(tester, &[]), Default::default(),),
         issue_417_1,
         "const test = process.env['x']",
         "const test = process.env['x']"
@@ -277,12 +327,11 @@ mod tests {
 
     test!(
         Default::default(),
-        |tester| as_folder(InlineGlobals {
-            envs: envs(tester, &[("x", "FOO")]),
-            globals: globals(tester, &[]),
-            typeofs: Default::default(),
-            bindings: Default::default()
-        }),
+        |tester| inline_globals(
+            envs(tester, &[("x", "FOO")]),
+            globals(tester, &[]),
+            Default::default(),
+        ),
         issue_417_2,
         "const test = process.env['x']",
         "const test = 'FOO'"
